#/usr/bin/env python3
"""
The Interface for running compile time sanity validates on an instal model
"""
##-- imports
from __future__ import annotations

import warnings
import abc
import logging as logmod
from collections import defaultdict
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from re import Pattern
from typing import (Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

logging = logmod.getLogger(__name__)

from jgdv.apis.clingo.ast import SolverAST

class SolverASTVisitor_i(metaclass=abc.ABCMeta):
    """
    Interface for the AST Visitor,
    the stub of whic which can be generated
    with instal.cli.generate_vistor
    """

    @abc.abstractmethod
    def visit(self, node:SolverAST, *, skip_actions=False): pass

    @abc.abstractmethod
    def visit_all(self, nodes:list[SolverAST]): pass

    @abc.abstractmethod
    def generic_visit(self, node): pass

    @abc.abstractmethod
    def flatten_for_classes(self, *classes): pass

    @abc.abstractmethod
    def add_actions(self, actions_obj): pass

@dataclass
class SolverValidationReport:
    """
    A Result of an InstalValidate.

    """
    ast       : SolverAST         = field()
    msg       : str               = field()
    level     : int               = field(kw_only=True)
    validator : Validator_i = field(kw_only=True)
    data      : Any               = field(kw_only=True, default=None)

    fmt       : None|str          = field(kw_only=True, default="({level}) {source}{loc} {msg}")

    def __repr__(self):
        return str(self) + f" AST: {self.ast}"

    def __str__(self):
        loc    = ""
        source = ""
        if self.ast and bool(self.ast.parse_source):
            source = f"Source: {self.ast.sources_str} : "

        if self.ast and self.ast.parse_loc is not None:
            loc = f"L:{self.ast.parse_loc[0]}, C:{self.ast.parse_loc[1]}: "

        return self.fmt.format_map({"msg"    : self.msg,
                                    "level"  : logmod.getLevelName(self.level),
                                    "source" : source,
                                    "loc"    : loc})

@dataclass
class Validator_i(metaclass=abc.ABCMeta):
    """
    The Core Interface for running compile time validates
    on an instal specification.

    parsers return list[SolverAST],

    the validation runner triggers an AST walk, which a validator will
    have registered actions on with `get_actions`

    then `validate` is called, and any data the validator's actions have stored
    will be used to generate reports

    NOTE: validators use an internal trio of debug/info/warning methods
    instead of just logging, or raising an error,
    so that *all* validates can be run, instead of throwing up to the runner on the first error.

    validator runners call `full_clear` at the start of a validation pass.
    which re-call's any validator's fields default_factories.

    """

    current_reports : list[SolverValidationReport] = field(init=False, default_factory=list)

    def clear(self):
        """
        an empty method for more involved reinitialization
        """
        pass

    def full_clear(self):
        """
        The full clear triggered by the validate runner
        """
        self.current_reports = []
        for field in self.__dataclass_fields__.values():
            if field.default_factory is None:
                continue

            setattr(self, field.name, field.default_factory())

        self.clear()

    def debug(self, msg, ast=None, data=None):
        self.build_note(ast, msg, logmod.DEBUG, data)

    def delay_info(self, msg, ast=None, data=None):
        self.build_note(ast, msg, logmod.INFO, data)

    def delay_warning(self, msg, ast=None, data=None):
        self.build_note(ast, msg, logmod.WARN, data)

    def delay_error(self, msg, ast=None, data=None):
        self.build_note(ast, msg, logmod.ERROR, data)

    def build_note(self, ast, msg, level, data):
        self.current_reports.append(SolverValidationReport(ast, msg,
                                                           level=level,
                                                           validator=self.__class__,
                                                           data=data))

    def __call__(self) -> list[SolverValidationReport]:
        """
        The access point used by InstalValidateRunner.
        Clears the log of reports generated, runs the .validate method,
        and returns the new list of reports.
        """
        self.validate()
        return self.current_reports[:]

    def validate(self): pass

@dataclass
class SolverValidatorRunner:
    """
    Given a collection of Instal Validators,
    and a list of SolverAST's,
    runs the validators on the ast's
    and warns / errors / reports a collection of results
    """

    validators : list[Validator_i]       = field(default_factory=list)
    visitor    : SolverASTVisitor_i      = field(default_factory=None)

    def __post_init__(self):
        # Register all validators' actions with the visitor
        for validator in self.validators:
            self.visitor.add_actions(validator)

        self.visitor.flatten_for_classes()

        logging.info("%s built with %s validators and visitor class %s",
                     self.__class__.__name__,
                     len(self.validators),
                     self.visitor.__class__.__name__)

    def validate(self, asts:list[SolverAST]) -> dict[int, list[SolverValidationReport]]:
        if not isinstance(asts, list):
            asts = [asts]
        logging.info("Running Validate on %s primary level asts", len(asts))
        total_results                = defaultdict(lambda: [])
        hard_fails : list[Exception] = []
        error_count : int            = 0

        for validator in self.validators:
            validator.clear()

        logging.debug("Visiting nodes")
        self.visitor.visit_all(asts)

        for validator in self.validators:
            logging.debug("Running Validator: %s", validator.__class__.__name__)
            # Run the Validate, recording results
            try:
                results = validator()
                # Collect the reports by level
                for note in results:
                    total_results[note.level].append(note)
                    error_count += 1 if note.level >= logmod.ERROR else 0

            except Exception as err:
                # If a validator actually *errors*, record that but keep going
                logging.exception("Validator Hard Failed: %s", validator)
                hard_fails.append(err)
                error_count += 1

        # When all validates are done, report exceptions
        if bool(error_count):
            just_errors = {x:y for x,y in total_results.items() if x >= logmod.ERROR}
            just_errors.update({101:hard_fails})
            raise Exception(f"Validating produced Errors: {error_count}", just_errors)

        # warning if theres any
        warnings = [report for x,y in total_results.items() for report in y if logmod.INFO < x <= logmod.ERROR]
        for report in sorted(warnings, key=lambda x:x.level):
            logging.warning(str(report))

        return dict(total_results)



"""
Exports a given trie to Netscape bookmark file format
Main function: exportBookmarks
"""
##-- imports
from __future__ import annotations

import logging

##-- end imports

from bs4 import BeautifulSoup

class NetscapeLoader:

    @staticmethod
    def read_netscape(path:pl.Path):
        logging.info('Starting html opener for: %s', path)
        with open(path, 'rb') as f:
            rawHtml = f.read().decode("utf-8","ignore")

        soup     = BeautifulSoup(rawHtml,'html.parser')
        bkmkList = NetscapeLoader._getLinks(soup)
        logging.info("Found %s links", len(bkmkList))
        return BookmarkCollection(bkmkList)

    @staticmethod
    def _getLinks(aSoup) -> List[Bookmark]:
        bkmks = aSoup.find_all('a')
        bkmkList = []
        for x in bkmks:
            tagString = x.get('tags')
            if tagString is not None:
                indTags = tagString.split(',')
            else:
                indTags = []
            tagSet = set(indTags)
            newBkmk = Bookmark(x.get('href'),tagSet, name=x.get_text())
            bkmkList.append(newBkmk)

        return bkmkList

class NetscapeWriter:

    groupCount = 0
    listCount  = 0
    entryCount = 0
    _header = """<!DOCTYPE NETSCAPE-Bookmark-file-1>
    <HTML>
    <META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=UTF-8">
    <Title>Bookmarks</Title>
    <H1>Bookmarks</H1>
    """
    _footer = "\n</HTML>"
    _group  = "<DT><H3 FOLDED> {} </H3> \n\t <DL><p> \n\t\t {} \n\t </DL><p>\n"
    _tags   = 'TAGS="{}"'
    _item   = '<DT><A HREF="{}" {}>{}</A>'

    @staticmethod
    def exportBookmarks(data:list|dict, target:pl.Path):
        """ Main function, returns a complete bookmark string to write to a file """
        writer = NetscapeWriter()

        result = []
        result.append(writer.header())
        results += _convertData(data)
        result.append(footer())
        logging.info("Finished converted bookmarks: %s groups | %s lists | %s entries", groupCount,listCount,entryCount)
        target.write_text(formatted_string)

    def header(self):
        """ Creates the Header for the entire bookmark file """
        return self.header

    def footer(self):
        """ Finishs the bookmark file """
        return self._footer

    def convertData(self, data) -> list[str]:
        results = []
        match data:
            case dict():
                self.groupCount  += 1
                keys         = sorted([x for x in data.keys()])
                results += [self.groupToNetscapeString(x,data[x]) for x in keys if isinstance(data[x],dict) and x != '__path']
                results += [self.bookmarkToItem(data[x]) for x in keys if isinstance(data[x], Bookmark) and x != '__path']
            case list():
                entryCount = len(data)
                results += [self.bookmarkToItem(x) for x in data]
            case _:
                raise Exception('Unrecognised conversion type: %s', type(data))

        return "\n".join(sorted(results))

    def groupToNetscapeString(self, name, data):
        return self._group.format(name,self.convertData(data))

    def bookmarkToItem(self, bkmk):
        """ Converts a single bookmark to html """
        assert(isinstance(bkmk, Bookmark))
        logging.debug("Exporting link: %s", bkmk.url)
        #add the link:
        tags = self._tags.format(",".join(sorted(bkmk.tags)))
        item = self._item.format(bkmk.url,tags, bkmk.name.replace('\n',''))
        return item



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class JGDVLogContext:
    """
      a really simple wrapper to set a logger's level, then roll it back
    """

    def __init__(self, logger, level=None):
        self._logger          = logger
        self._original_level  = self._logger.level
        self._level_stack     = [self._original_level]
        self._temp_level      = level or self._original_level

    def __call__(self, level):
        self._temp_level = level
        return self

    def __enter__(self):
        match self._temp_level:
            case int() | str():
                self._level_stack.append(self._logger.level)
                self._logger.setLevel(self._temp_level)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> bool:
        if bool(self._level_stack):
            self._logger.setLevel(self._level_stack.pop())
        else:
            self._logger.setLevel(self._original_level)


    def __getattr__(self, key):
        return getattr(self._logger, key)



"""
The parser for the REPL

"""
##-- imports
from __future__ import annotations

import logging as logmod

##-- end imports

import pyparsing as pp

logging = logmod.getLogger(__name__)
config  = acab.config

rst     = pp.delimited_list(pp.rest_of_line, delim=pp.White("\n\r"), combine=True).leave_whitespace()

##-- multi line
MULTI_LINE_START = config.module.REPL.MULTI_LINE_START
MULTI_LINE_END   = config.module.REPL.MULTI_LINE_END

multi_line_start = pp.Regex(MULTI_LINE_START)
multi_line_end = pp.Regex(MULTI_LINE_END)

multi_line_start.set_parse_action(lambda s,l,t: "multi")
multi_line_end.set_parse_action(lambda s,l,t: "multi")

##-- end multi line

shortcut_config   = config.module.REPL.shortcuts
short_cmd_parsers = []

##-- shortcuts

def gen_action(cmd):

    def __action(s, l, t):
        return cmd

    return __action

for cmd in shortcut_config._keys:
    kw      = shortcut_config[cmd]
    s_cmd_p = pp.Keyword(kw)
    s_cmd_p.set_parse_action(gen_action(cmd))
    short_cmd_parsers.append(s_cmd_p)

##-- end shortcuts

sugared = pp.Suppress(pp.Literal(":")) + pp.MatchFirst(short_cmd_parsers) + rst

precmd_parser = pp.MatchFirst([multi_line_start,
                               multi_line_end + rst,
                               sugared,
                               rst]).leave_whitespace()

##-- step kws
back_kw     = pp.Keyword("back")
rule_kw     = pp.Keyword("rule")
layer_kw    = pp.Keyword("layer")
pipe_kw     = pp.Keyword("pipe")
pipeline_kw = pp.Keyword("pipeline")
step_parser = pp.MatchFirst([back_kw,
                             rule_kw + rst,
                             layer_kw,
                             pipe_kw,
                             pipeline_kw])

##-- end step kws

##-- stat kws
operator_kw  = pp.MatchFirst([pp.Keyword("ops"),
                              pp.Keyword("operator")])("operator")
module_kw    = pp.MatchFirst([pp.Keyword("mod"),
                              pp.Keyword("module")])("module")
semantic_kw  = pp.Keyword("semantics")("semantics")
printer_kw   = pp.Keyword("printers")("printers")

stats_parser = pp.ZeroOrMore(pp.MatchFirst([operator_kw,
                                            module_kw,
                                            semantic_kw,
                                            printer_kw]))

##-- end stat kws

##-- listener
# add/remove/list/threshold
add_kw        = pp.Keyword("add")
remove_kw     = pp.Keyword("remove")
list_kw       = pp.Keyword("list")

listen_parser = pp.Or([add_kw, remove_kw, list_kw])

##-- end listener

##-- parser debug
signals_kw        = pp.Keyword("signals")("signals")
sugar_kw          = pp.Keyword("sugar")("sugar")
debug_kw          = pp.Suppress(pp.Keyword("debug"))
disable_kw        = pp.Keyword("disable")

debug_control     = debug_kw + pp.Or([disable_kw, list_kw, rst])("debug")
parse_info_parser = signals_kw | sugar_kw | debug_control | rst

##-- end parser debug

##-- contexts
number = pp.common.integer

slice_p = (PU.s(pp.Literal('[')) +
           PU.op(number)('first') +
           PU.op(pp.Literal(':')("mid") +
                 PU.op(number)("second")) +
           PU.s(pp.Literal(']')))
slice_p.set_parse_action(build_slice)

ctx_kw    = pp.MatchFirst([pp.Keyword("ctx"),
                           pp.Keyword("c")])("context")
ctx       = number("short_context")
ctx_slice = slice_p("context_slice")

# $x, @x
binding = AP.VALBIND

# ctx ([2:])? $x?
ctx_parser = (ctx_kw + pp.Optional(ctx | ctx_slice)
              + pp.ZeroOrMore(binding)("bindings"))

##-- end contexts

##-- context selection

ctx_index  = number("subset")
ctx_subset = slice_p("subset")
clear_kw   = pp.Keyword("clear")("clear")
minus_kw   = pp.Keyword("-")("clear")

ctx_select_parser = pp.MatchFirst([ctx_subset,
                                   ctx_index,
                                   clear_kw,
                                   minus_kw,
                                   rst])

##-- end context selection



"""
Events express when a value holds in time
"""
##-- imports
from __future__ import annotations
from .arc import Arc
from fractions import Fraction
import logging as root_logger

##-- end imports

logging = root_logger.getLogger(__name__)

class Event:
    """ A Value active during a timespan """

    def __init__(self, a, b, value_is_pattern=False, params=None):
        assert(isinstance(a, Arc))
        self.arc              = a.copy()
        self.values           = b
        self.parameters       = params or {}
        self.value_is_pattern = value_is_pattern

    def __call__(self, count, just_values=False, rand_s=None):
        """ Get a list of events given a time """
        match (count in self.arc), self.value_is_pattern:
            case True, True:
                return self.values(count - self.arc.start, just_values)
            case False, True:
                return [self]
            case _:
                return []

    def __contains__(self, other):
        return other in self.arc

    def __repr__(self):
        return f"{self.values} :: {self.arc}"


    def base(self):
        """ Get all fractions used in this event """
        time_list = self.arc.pair()
        if self.value_is_pattern:
            time_list += [x - self.arc.start for x in self.values.base()]
        return set(time_list)

    def key(self):
        """ Get the start of the event, for sorting """
        return self.arc.start

    def print_flip(self, start=True):
        """ Get a string describing the event's entry/exit status """
        return f"⤒{self.values} " if start else f"{self.values}⤓"

    def __getitem__(self, val):
        """ event[x] """
        return self.parameters[val]
